#| -*-Scheme-*-

$Id$

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,
USA.

|#

;;;; Operations on Code in Locales

(declare (usual-integrations))

(define (locale-eval expression locale)
  (locale-scode-eval (syntax expression locale) locale))

(define (locale-scode-eval scode locale)
  (extended-scode-eval scode (locale/environment locale)))

(define (locale-load pathname locale)
  (fluid-let ((hook/repl-eval
	       (lambda (expression locale repl)
		 (repl-scode-eval (syntax expression locale)
				  (locale/environment locale)
				  repl))))
    (load pathname (locale/environment locale))))

(define (syntax-locale locale)
  (let ((undefined '()))
    (fluid-let ((hook/note-undefined
		 (lambda (name)
		   (set! undefined (cons name undefined))
		   unspecific)))
      ((lambda (scode)
	 (if (pair? undefined)
	     (warn "References to undefined variables:" undefined))
	 scode)
       (syntax* (let ((source-forms (locale/source-forms locale))
		      (declarations (locale/declarations locale)))
		  (if (pair? declarations)
		      `((,(close-syntax 'DECLARE system-global-environment)
			 ,@declarations)
			,@source-forms)
		      source-forms))
		locale)))))

(define (default/note-undefined identifier)
  (warn "Reference to undefined variable:" identifier))

(define hook/note-undefined default/note-undefined)

(define (note-undefined identifier)
  (hook/note-undefined identifier))

(define (locale/source-forms locale)
  (let ((defaults (locale/source-pathname-defaults locale)))
    (let loop ((clauses (locale/clauses locale))
	       (forms '()))
      (if (null? clauses)
	  (reverse! forms)
	  (loop (cdr clauses)
		(let ((clause (car clauses)))
		  (case (car clause)
		    ((FILES)
		     (fold (lambda (pathname forms)
			     (append-reverse
			      (read-forms-from-file
			       (merge-pathnames pathname defaults))
			      forms))
			   forms
			   (cdr clause)))
		    ((CODE)
		     (append-reverse (cdr clause) forms))
		    (else forms))))))))

;;; This perhaps ought to permit the locale to supply its own reader,
;;; or at least parser table, with arbitrary lexical extensions.

(define (read-forms-from-file pathname)
  (call-with-input-file pathname
    (lambda (input-port)
      (let loop ((forms '()))
	(let ((form (read input-port)))
	  (if (eof-object? form)
	      (reverse! forms)
	      (loop (cons form forms))))))))

(define (locale/declarations locale)
  ;++ This is bogus.
  sf/default-declarations)
