#| -*-Scheme-*-

$Id$

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,
USA.

|#

;;;; Dumping and Loading Locale Information

(declare (usual-integrations))

(define (locale-fasdump scode pathname locale)
  (fasdump (externalize-references scode locale)
	   (pathname-default-type pathname "bin")))

(define (locale-fasload pathname locale)
  (internalize-references (fasload pathname) locale))

(define (fasdump-locale-bindings locale #!optional pathname)
  (cond ((locale/bindings-pathname locale pathname)
	 => (lambda (pathname)
	      (fasdump
	       (map (lambda (identifier.binding)
		      (let ((identifier (car identifier.binding))
			    (binding (cdr identifier.binding)))
			(vector (qualify identifier locale)
				(binding/signature binding)
				(externalize-item (binding/static binding)
						  locale))))
		    (hash-table->alist (locale/bindings locale)))
	       pathname)))
	(else
	 (error "Don't know where to dump locale bindings:" locale))))

(define (fasdump-locale-scode locale #!optional pathname)
  (cond ((locale/scode-pathname locale pathname)
	 => (lambda (pathname)
	      (locale-fasdump (locale/scode locale) pathname locale)))
	(else
	 (error "Don't know where to dump locale scode:" locale))))

(define (maybe-fasload-locale-bindings locale #!optional pathname)
  (cond ((locale/bindings-pathname locale pathname)
	 => (lambda (pathname)
	      (call-with-current-continuation
		(lambda (k)
		  (for-each
		   (lambda (binding)
		     (locale/define locale
				    (unqualify (vector-ref binding 0) locale)
				    (vector-ref binding 1)
				    (internalize-item (vector-ref binding 2)
						      locale)))
		   (bind-condition-handler (list condition-type:error)
		       (lambda (condition) condition (k #f))
		     (lambda ()
		       (fasload pathname))))))
	      #t))
	(else #f)))

(define (maybe-fasload-locale-scode locale #!optional pathname)
  (cond ((locale/scode-pathname locale pathname)
	 => (lambda (pathname)
	      (call-with-current-continuation
		(lambda (k)
		  ((lambda (scode)
		     ;; This is random.  We choose the unoptimized
		     ;; scode because we don't have the external
		     ;; integration information.
		     (set-locale/unoptimized-scode! locale scode)
		     scode)
		   (bind-condition-handler (list condition-type:error)
		       (lambda (condition) condition (k #f))
		     (lambda ()
		       (locale-fasload pathname locale))))))))
	(else #f)))

(define (make-reference-mapping-walker comment-handler)
  (make-scode-walker
   (lambda (expression locale walker)
     locale walker			;ignore
     expression)
   `((ACCESS ,map-references/access)
     (ASSIGNMENT ,map-references/assignment)
     (COMBINATION ,map-references/combination)
     (COMMENT ,comment-handler)
     (CONDITIONAL ,map-references/conditional)
     (DECLARATION ,map-references/declaration)
     (DEFINITION ,map-references/definition)
     (DELAY ,map-references/delay)
     (DISJUNCTION ,map-references/disjunction)
     (LAMBDA ,map-references/lambda)
     (SEQUENCE ,map-references/sequence))))

(define (map-references scode locale walker)
  ((scode-walk walker scode) scode locale walker))

(define (map-references/access access locale walker)
  (make-access
   (map-references (access-environment access) locale walker)
   (access-name access)))

(define (map-references/assignment assignment locale walker)
  (make-assignment
   (assignment-name assignment)
   (map-references (assignment-value assignment) locale walker)))

(define (map-references/combination combination locale walker)
  (make-combination
   (map-references (combination-operator combination) locale walker)
   (map (lambda (operand)
	  (map-references operand locale walker))
	(combination-operands combination))))

(define (map-references/conditional conditional locale walker)
  (make-conditional
   (map-references (conditional-predicate conditional) locale walker)
   (map-references (conditional-consequent conditional) locale walker)
   (map-references (conditional-alternative conditional) locale walker)))

(define (map-references/declaration declaration locale walker)
  (make-declaration
   (declaration-text declaration)
   (map-references (declaration-expression declaration) locale walker)))

(define (map-references/definition definition locale walker)
  (make-definition
   (definition-name definition)
   (map-references (definition-value definition) locale walker)))

(define (map-references/delay delay locale walker)
  (make-delay
   (map-references (delay-expression delay) locale walker)))

(define (map-references/disjunction disjunction locale walker)
  (make-disjunction
   (map-references (disjunction-predicate disjunction) locale walker)
   (map-references (disjunction-alternative disjunction) locale walker)))

(define (map-references/lambda *lambda locale walker)
  (lambda-components** *lambda
    (lambda (pattern bound body)
      (make-lambda** pattern bound (map-references body locale walker)))))

(define (map-references/sequence sequence locale walker)
  (make-sequence
   (map (lambda (action)
	  (map-references action locale walker))
	(sequence-actions sequence))))

(define (externalize-references scode locale)
  (map-references scode locale reference-externalizers))

(define reference-externalizers
  (make-reference-mapping-walker
   (lambda (comment locale walker)
     (externalize-references/comment comment locale walker))))

(define (externalize-references/comment comment locale walker)
  (let ((text (comment-text comment))
	(expression (comment-expression comment)))
    (if (and (identifier? text)
	     (access? expression)
	     (interpreter-environment? (access-environment expression)))
	(let ((qualified (qualify text locale)))
	  (make-comment qualified (make-unresolved-external-error qualified)))
	(make-comment text (map-references expression locale walker)))))

(define (internalize-references scode locale)
  (map-references scode locale reference-internalizers))

(define reference-internalizers
  (make-reference-mapping-walker
   (lambda (comment locale walker)
     (internalize-references/comment comment locale walker))))

(define (internalize-references/comment comment locale walker)
  (let ((text (comment-text comment))
	(expression (comment-expression comment)))
    (if (and (or (symbol? text)
		 (qualified? text))
	     (unresolved-external-error? expression))
	(let ((identifier (unqualify text locale)))
	  (make-comment identifier
			(cond ((lookup-identifier locale identifier)
			       => (lambda (item)
				    (binding-access
				     (top-level-variable-item/binding item))))
			      (else
			       (warn "Unresolved external:" text locale)
			       expression))))
	(make-comment text (map-references expression locale walker)))))

(define (externalize-item item locale)
  (and item
       (if (not (keyword-value-item? item))
	   (error "Undumpable item:" item)
	   (let ((source (keyword-value-item/source item)))
	     (if (identifier? source)
		 (qualify source locale)
		 (cons (car source)
		       (externalize-references (cdr source) locale)))))))

(define (internalize-item item locale)
  (and item
       (if (or (symbol? item) (qualified? item))
	   (locale/lookup locale (unqualify item locale))
	   (let ((transformer-type (car item))
		 (scode (internalize-references (cdr item) locale)))
	     (make-keyword-value-item
	      ((case transformer-type
		 ((SC-MACRO-TRANSFORMER)
		  sc-macro-transformer->expander)
		 ((RSC-MACRO-TRANSFORMER)
		  rsc-macro-transformer->expander)
		 ((ER-MACRO-TRANSFORMER)
		  er-macro-transformer->expander)
		 ((NON-HYGIENIC-MACRO-TRANSFORMER)
		  non-hygienic-macro-transformer->expander))
	       (transformer-eval scode locale)
	       locale)
	      (cons transformer-type scode))))))

(define unresolved-external-error-message "Unresolved external:")

(define (make-unresolved-external-error qualified)
  (make-combination
   (make-primitive-procedure 'ERROR-PROCEDURE)
   (list unresolved-external-error-message
	 (list qualified)
	 (make-the-environment))))

(define (unresolved-external-error? expression)
  (and (combination? expression)
       (eq? (combination-operator expression)
	    (make-primitive-procedure 'ERROR-PROCEDURE))
       (let ((operands (combination-operands expression)))
	 (and (= 3 (length operands))
	      (string=? unresolved-external-error-message (car operands))))))
