#| -*-Scheme-*-

$Id$

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

|#

;;;; Interfaces

(declare (usual-integrations))

(define-structure (interface
                   (constructor make-interface
                                (method/ref method/walk))
                   (conc-name interface/))
  (method/ref #f read-only #t)
  (method/walk #f read-only #t)
  (clients (make-population) read-only #t)
  )

(define (note-reference-to-interface! interface client)
  (add-to-population! (interface/clients interface) client))

(define (forget-reference-to-interface! interface client)
  (remove-from-population! (interface/clients interface) client))

(define (interface-ref interface name)
  ((interface/method/ref interface) name))

(define (for-each-declaration procedure interface)
  ((interface/method/walk interface) procedure))

;;;; Simple & Compound Interfaces

(define (make-simple-interface export-specifiers)
  (let ((export-table (process-simple-exports export-specifiers)))
    (make-interface
     (lambda (name)
       (cond ((hash-table/get export-table name #f)
              => (lambda (signature)
                   (values name signature)))
             (else (values #f #f))))
     (lambda (procedure)
       (hash-table/for-each export-table
                            (lambda (name signature)
                              (procedure name name signature)))))))

(define (process-simple-exports export-specifiers)
  (let ((export-table (make-symbol-hash-table)))
    (define (enter name signature)
      (hash-table/put! export-table name signature))
    (for-each (lambda (specifier)
                (define (lose)
                  (error "Invalid export specifier:" specifier))
                (cond ((symbol? specifier)
                       (enter specifier undeclared-signature))
                      ((and (pair? specifier)
                            (pair? (cdr specifier))
                            (null? (cddr specifier)))
                       (let ((name (car specifier))
                             (signature
                              (process-export-signature
                               (cadr specifier))))
                         (cond ((symbol? name)
                                (enter name signature))
                               ((list-of-type? name symbol?)
                                (for-each (lambda (name)
                                            (enter name signature))
                                          name))
                               (else (lose)))))
                      (else (lose))))
                export-specifiers)
    export-table))

(define (process-export-signature signature)
  signature)

(define (make-compound-interface . component-thunks)
  (let ((interface (make-interface
                    (lambda (name)
                      (let loop ((component-thunks component-thunks))
                        (if (pair? component-thunks)
                            (receive (internal-name signature)
                                (interface-ref ((car component-thunks))
                                               name)
                              (if internal-name
                                  (values internal-name signature)
                                  (loop (cdr component-thunks))))
                            (values #f #f))))
                    (lambda (procedure)
                      (for-each (lambda (component-thunk)
                                  (for-each-declaration
                                   procedure
                                   (component-thunk)))
                                component-thunks)))))
    (for-each (lambda (component-thunk)
                (note-reference-to-interface! (component-thunk)
                                              interface))
              components)
    interface))

;;;; Modified Interfaces

;;; This uses a much more straightforward implementation technique
;;; than Scheme48's.  The difference in performance is probably
;;; negligible, but this is much easier to understand.

(define (make-modified-interface-maker commands)
  (if (null? commands)
      (lambda (interface) interface)
      (let ((translation (make-translation commands)))
        (let ((ref-maker (make-modified-ref-maker translation))
              (walk-maker (make-modified-walk-maker translation)))
          (lambda (interface)
            (let ((interface*
                   (make-interface (ref-maker interface)
                                   (walk-maker interface))))
              (note-reference-to-interface! interface interface*)
              interface*))))))

(define (make-modified-ref-maker translation)
  (let ((mapper (translation/mapper translation)))
    (lambda (interface)
      (lambda (name)
        (interface-ref interface (mapper name))))))

(define (make-modified-walk-maker translation)
  (let ((walker (translation/walker translation)))
    (lambda (interface)
      (lambda (procedure)
        (walker (lambda (walk-aliases)
                  (for-each-declaration
                   (lambda (external internal signature)
                     (walk-aliases external
                                   (lambda (external*)
                                     (procedure external*
                                                internal
                                                signature))))
                   interface))
                (lambda (external intermediate)
                  (receive (internal signature)
                      (interface-ref interface intermediate)
                    (if internal
                        (procedure external internal signature)))))))))
