#| -*-Scheme-*-

$Id$

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301,
USA.

|#

;;;; Locales, Bindings, and Implementations

(declare (usual-integrations))

(define-structure (locale
                   (constructor make-locale
                                (imports-thunk
                                 syntactic-tower
                                 source-pathname-defaults
				 object-pathname-defaults
                                 clauses))
                   (conc-name locale/))
  imports-thunk
  (%imports #f)
  (environment (make-root-top-level-environment))
  (bindings (make-eq-hash-table) read-only #t)
  syntactic-tower
  (source-pathname-defaults #f read-only #t)
  (object-pathname-defaults #f)
  clauses
  (clients (make-population) read-only #t)
  (early-bound? #f)
  (unoptimized-scode #f)
  (optimized-scode #f)			;scode
  (externs-block #f)			; optimizer
  (externs #f)				;  cruft
  (loaded? #f)
  (%qualified-cache #f)
  )

(define (locale/imports locale)
  (maybe-initialize-locale! locale)
  (locale/%imports locale))

(define (initialize-locale! locale)
  (let ((imports ((locale/imports-thunk locale))))
    (set-locale/%imports! locale imports)
    (for-each (lambda (import)
                (note-reference-to-implementation! import locale))
              imports)))

(define (maybe-initialize-locale! locale)
  (if (not (locale/%imports locale))
      (initialize-locale! locale)))

(define (reinitialize-locale! locale)
  (cond ((locale/%imports locale)
         => (lambda (imports)
              (for-each (lambda (import)
                          (forget-reference-to-implementation! import locale))
                        imports))))
  (initialize-locale! locale))

(define (locale/scode locale)
  (or (locale/optimized-scode locale)
      (locale/unoptimized-scode locale)))

(define (object-pathname-accessor type)
  (lambda (locale #!optional pathname)
    (cond ((or (and (not (default-object? pathname))
		    pathname)
	       (locale/object-pathname-defaults locale))
	   => (lambda (pathname)
		(pathname-new-type pathname type)))
	  (else #f))))

(define locale/bindings-pathname (object-pathname-accessor "bbn"))
(define locale/scode-pathname (object-pathname-accessor "bin"))
(define locale/externs-pathname (object-pathname-accessor "ext"))
(define locale/compiled-pathname (object-pathname-accessor "com"))
(define locale/info-pathname (object-pathname-accessor "bci"))

(define (note-reference-to-locale! locale client)
  (add-to-population! (locale/clients locale) client))

(define (forget-reference-to-locale! locale client)
  (remove-from-population! (locale/clients locale) client))

(define (locale/lookup locale identifier)
  (cond ((hash-table/get (locale/bindings locale) identifier #f))
        ((locale/lookup-imported locale identifier)
         => (lambda (binding)
              (hook/locale/lookup-imported locale identifier binding)
              (import-binding binding
                              (locale/environment locale)
                              identifier)
              binding))
        (else #f)))

;;; This hook might be used for certain kinds of cross-referencing.
;;; I'm not sure whether it's really useful.

(define (default/locale/lookup-imported locale identifier binding)
  locale identifier binding             ;ignore
  unspecific)

(define hook/locale/lookup-imported default/locale/lookup-imported)

(define (locale/lookup-imported locale identifier)
  (let loop ((imports (locale/imports locale)))
    (and (pair? imports)
         (or (implementation/lookup (car imports) identifier)
             (loop (cdr imports))))))

(define (locale/define locale identifier signature static)
  (let ((bindings (locale/bindings locale)))
    (cond ((hash-table/get bindings identifier #f)
           => (lambda (binding)
                (clobber-binding! binding signature static)))
          (else
           (hash-table/put!
            bindings
            identifier
            (make-binding (locale/environment locale)
                          identifier
                          signature
                          static))))))

(define-structure (binding
                   (constructor %make-binding
                                (%environment
                                 identifier
                                 signature
                                 static))
                   (conc-name binding/))
  (%environment #f read-only #t)
  (identifier #f read-only #t)
  (signature #f)
  (static #f))

(define (make-binding environment identifier signature static)
  (let ((binding
         (%make-binding environment identifier signature static)))
    (local-assignment
     (binding/environment binding)
     identifier                         ;++ Make sure it's a symbol!
     (cond ((variable-signature? signature)
            (make-unmapped-unassigned-reference-trap))
           ((syntactic-signature? signature)
            (make-unmapped-macro-reference-trap static))
           (else
            (error "Invalid binding:" identifier signature static))))
    binding))

(define (binding/environment binding)
  (->environment (binding/%environment binding)))

(define (clobber-binding! binding signature static)
  (let ((environment (binding/environment binding))
        (identifier (binding/identifier binding)))
    (if (eq? (environment-reference-type environment identifier)
             'MACRO)
        (local-assignment
         environment
         identifier
         (or static
             (make-unmapped-unassigned-reference-trap)))))
  (set-binding/signature! binding signature)
  (set-binding/static! binding static))

(define (binding-access binding)
  (make-access (binding/environment binding) (binding/identifier binding)))

(define (import-binding binding target-environment target-identifier)
  (link-variables target-environment
                  target-identifier
                  (binding/environment binding)
                  (binding/identifier binding)))

(define (binding-with-signature binding signature)
  (%make-binding (binding/environment binding)
                 (binding/identifier binding)
                 signature
                 (binding/static binding)))

(define (binding-sans-integration binding)
  (if (variable-signature? (binding-signature binding))
      (%make-binding (binding/environment binding)
                     (binding/identifier binding)
                     (binding/signature binding)
                     #f)
      binding))

(define-structure (implementation
                   (constructor %make-implementation
                                (interface-thunk
                                 translation
                                 method/lookup
                                 method/->locale))
                   (conc-name implementation/))
  interface-thunk
  (%interface #f)
  (translation #f read-only #t)
  (method/lookup #f read-only #t)
  (method/->locale #f read-only #t)
  (clients (make-population) read-only #t)
  )

(define (implementation/interface implementation)
  (maybe-initialize-implementation! implementation)
  (implementation/%interface implementation))

(define (initialize-implementation! implementation)
  (let ((interface ((implementation/interface-thunk implementation))))
    (set-implementation/%interface! implementation interface)
    (note-reference-to-interface! interface implementation)))

(define (maybe-initialize-implementation! implementation)
  (if (not (implementation/%interface implementation))
      (initialize-implementation! implementation)))

(define (reinitialize-implementation! implementation)
  (cond ((implementation/%interface implementation)
         => (lambda (interface)
              (forget-reference-to-interface! interface implementation))))
  (initialize-implementation! implementation))

(define (implementation/locale implementation)          ;++ bogus
  ((implementation/method/->locale implementation)))

(define (note-reference-to-implementation! implementation client)
  (add-to-population! (implementation/clients implementation) client))

(define (forget-reference-to-implementation! implementation client)
  (remove-from-population! (implementation/clients implementation) client))

(define (implementation/lookup implementation identifier)
  (let ((interface (implementation/interface implementation)))
    (receive (internal-name signature) (interface-ref interface identifier)
      (cond ((and internal-name
                  ((implementation/method/lookup implementation)
                   (translate-name (implementation/translation implementation)
				   internal-name)))
             => (lambda (binding)
                  (receive (compatible? new-signature)
                      (match-signatures (binding/signature binding) signature)
                    (if (not compatible?)
                        (warn "Incompatible signature of" identifier
                              (error-irritant/noise " in") implementation
                              (error-irritant/noise " -- defined as")
                              (binding/signature binding)
                              (error-irritant/noise " but exported as")
                              signature))
                    (binding-with-signature binding new-signature))))
            (else #f)))))

(define (make-implementation interface-thunk translation
          method/lookup
          #!optional method/->locale)
  (%make-implementation interface-thunk
                        translation
                        method/lookup
                        (if (default-object? method/->locale)
                            (lambda () #f)
                            method/->locale)))

(define (make-implementation-from-locale locale interface-thunk
                                         #!optional translation)
  (let ((implementation
         (make-implementation interface-thunk
             (if (default-object? translation)
                 null-translation
                 translation)
           (lambda (name)
             (locale/lookup locale name))
           (lambda () locale))))
    (note-reference-to-locale! locale implementation)
    implementation))

(define (make-implementation-from-environment environment
                                              interface-thunk
                                              #!optional translation)
  (make-implementation interface-thunk
      (if (default-object? translation)
          null-translation
          translation)
    (lambda (name)
      (case (environment-reference-type environment name)
        ((UNBOUND) #f)
        ((UNASSIGNED NORMAL)
         (%make-binding environment name vanilla-variable-signature #f))
        ((MACRO)
         (%make-binding environment
                        name
                        syntactic-signature
                        (environment-lookup-macro environment name)))))))

(define (make-modified-implementation implementation commands)
  (let* ((interface-maker (make-modified-interface-maker commands))
         (implementation*
          (make-implementation
              (lambda ()
                (interface-maker
                 (implementation/interface implementation)))
              (implementation/translation implementation)
            (lambda (name)
              (implementation/lookup implementation name))
            (lambda ()
              (implementation/locale implementation)))))
    (note-reference-to-implementation! implementation implementation*)
    implementation*))

;;; Edwin Variables:
;;; lisp-indent/make-implementation: 2
;;; End:
