#| -*-Scheme-*-

$Id$

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,
USA.

|#

;;;; Locale Phases and Dependencies

(declare (usual-integrations))

;;; It would be nice to have some sort of real dependency management
;;; mechanism.  For now, though, this will work.

(define (locale/phase/early-bind locale)
  (if (not (locale/early-bound? locale))
      (begin
	(if (not (maybe-fasload-locale-bindings locale))
	    (if (locale/scode locale)
		(error "Locale is syntaxed but not early-bound:" local)
		(locale/phase/syntax locale)))
	(set-locale/early-bound?! locale #t))))

(define (locale/phase/syntax locale)
  (or (locale/scode locale)
      (maybe-fasload-locale-scode locale)
      (begin
	(for-each (lambda (import)
		    (cond ((implementation/locale import)
			   => locale/phase/early-bind)))
		  (locale/imports locale))
	(let ((scode (syntax-locale locale)))
	  (set-locale/unoptimized-scode! locale scode)
	  (set-locale/early-bound?! locale #t)
	  scode))))

;;; At some point, we ought to do something with the externs block and
;;; externs.

(define (locale/phase/optimize locale)
  (or (locale/optimized-scode locale)
      (maybe-fasload-locale-scode locale)
      (integrate/scode (syntax-locale locale)
        (lambda (optimized-scode externs-block externs)
	  (set-locale/unoptimized-scode! locale #f)
	  (set-locale/optimized-scode! locale optimized-scode)
	  (set-locale/externs-block! locale externs-block)
	  (set-locale/externs! locale externs)
	  (set-locale/early-bound?! locale #t)
	  optimized-scode))))

(define (locale/phase/load locale)
  (if (not (locale/loaded? locale))
      (begin
	(for-each (lambda (import)
		    (cond ((implementation/locale import)
			   => locale/phase/load)))
		  (locale/imports locale))
	(locale/phase/early-bind locale)
	(let ((value (locale-scode-eval (locale/phase/syntax locale) locale)))
	  (set-locale/loaded?! locale #t)
	  value))))

(define (walk-locale-graph roots procedure)
  (let ((visited (make-eq-hash-table)))
    (define (walk* roots)
      (for-each walk roots))
    (define (walk root)
      (if (not (hash-table/get visited root #f))
	  (begin
	    (hash-table/put! visited root #t)
	    (cond ((implementation? root)
		   (cond ((implementation/locale root)
			  => walk-locale)))
		  ((locale? root)
		   (walk-locale root))))))
    (define (walk-locale locale)
      (procedure locale)
      (walk* (locale/imports locale)))
    (walk* roots)))

(define (collect-locales roots filter)
  (let ((locales '()))
    (walk-locale-graph roots
      (lambda (locale)
	(if (filter locale)
	    (set! locales (cons locale locales)))))
    (reverse! locales)))

;;; Utility for building makefile rules.

(define (locale-graph->source-pathnames roots)
  (let ((pathnames '()))
    (walk-locale-graph roots
      (lambda (locale)
	(let ((defaults (locale/source-pathname-defaults locale)))
	  (for-each (lambda (clause)
		      (if (eq? 'FILES (car clause))
			  (set! pathnames
				(reverse-map/tail
				 pathnames
				 (lambda (pathname)
				   (merge-pathname pathname defaults))
				 (cdr clause)))))
		    (locale/clauses locale)))))
    (reverse! pathnames)))

(define (reverse-map/tail tail procedure list)
  (let loop ((list list) (tail tail))
    (if (pair? list)
	(loop (cdr list)
	      (cons (procedure (car list))
		    tail))
	tail)))
