#| -*-Scheme-*-

$Id$

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

|#

;;;; Qualified Names

(declare (usual-integrations))

(define-structure (qualified
		   (type vector)
		   (named '>>))		;Scheme48 convention
  name
  introducer
  uid)

(define (qualify identifier locale)
  (if (not (synthetic-identifier? identifier))
      identifier
      (let ((name (syntactic-closure/form identifier))
	    (uid (synthetic-identifier->uid identifier))
	    (environment (syntactic-closure/environment identifier)))
	(cond ((locale? environment)
	       (if (eq? environment locale)
		   identifier
		   (error "Non-local locale reference:"
			  `(QUALIFY ,identifier ,locale))))
	      ((syntactic-environment/introducer environment identifier)
	       => (lambda (introducer)
		    (make-qualified (qualify name locale)
				    (qualify introducer locale)
				    uid)))
	      (else
	       (error "Unable to qualify synthetic identifier:"
		      `(QUALIFY ,identifier ,locale)))))))

(define (unqualify qualified locale)
  (if (not (qualified? qualified))
      qualified
      (hash-table/intern! (locale/qualified-cache locale) qualified
	(lambda ()
	  (make-syntactic-closure
	   (expander-item/environment
	    (strip-keyword-value-item
	     (binding/static
	      (locale/lookup locale
			     (unqualify (qualified-introducer qualified)
					locale)))))
	   '()
	   (unqualify (qualified-name qualified) locale))))))

(define (locale/qualified-cache locale)
  (or (locale/%qualified-cache locale)
      (let ((cache (make-qualified-cache)))
	(set-locale/%qualified-cache! locale cache)
	cache)))

(define (make-qualified-cache)
  (make-hash-table qualified=? qualified-hash))

(define (qualified-hash qualified modulus)
  (remainder (qualified-uid qualified) modulus))

(define (qualified=? qualified-a qualified-b)
  (cond ((and (symbol? qualified-a) (symbol? qualified-b))
	 (eq? qualified-a qualified-b))
	((or (symbol? qualified-a) (symbol? qualified-b))
	 #f)
	(else
	 (and (qualified=? (qualified-name qualified-a)
			   (qualified-name qualified-b))
	      (qualified=? (qualified-introducer qualified-a)
			   (qualified-introducer qualified-b))
	      (eqv? (qualified-uid qualified-a)
		    (qualified-uid qualified-b))))))

(define synthetic-identifier-uids (make-eq-hash-table))

(define *synthetic-identifier-uid* 0)

(define (synthetic-identifier->uid identifier)
  (hash-table/intern! synthetic-identifier-uids identifier
    (lambda ()
      (set! *synthetic-identifier-uid* (+ 1 *synthetic-identifier-uid*)))))
