(define scheme-interface
  (make-simple-interface
   '(
     *
     +
     -
     /
     <
     <=
     =
     >
     >=
     abs
     acos
     (and :syntax)
     angle
     append
     apply
     asin
     assoc
     assq
     assv
     atan
     (begin :syntax)
     boolean?
     car
     cdr
     caar
     cadr
     cdar
     cddr
     caaar
     caadr
     cadar
     caddr
     cdaar
     cdadr
     cddar
     cdddr
     caaaar
     caaadr
     caadar
     caaddr
     cadaar
     cadadr
     caddar
     cadddr
     cdaaar
     cdaadr
     cdadar
     cdaddr
     cddaar
     cddadr
     cdddar
     cddddr
     call-with-current-continuation
     call-with-input-file
     call-with-output-file
     call-with-values
     (case :syntax)
     ceiling
     char->integer
     char-alphabetic?
     char-ci<=?
     char-ci<?
     char-ci=?
     char-ci>=?
     char-downcase
     char-lower-case
     char-numeric?
     char-ready?
     char-upcase
     char-upper-case?
     char-whitespace?
     char<=?
     char<?
     char=?
     char>=?
     char>?
     char?
     close-input-port
     close-output-port
     complex?
     (cond :syntax)
     cons
     cos
     current-input-port
     current-output-port
     (define :syntax)
     (define-syntax :syntax)
     (delay :syntax)
     denominator
     display
     (do :syntax)
     dynamic-wind
     eof-object?
     eq?
     equal?
     eqv?
     eval
     even?
     exact->inexact
     exact?
     exp
     expt
     floor
     for-each
     force
     gcd
     (if :syntax)
     imag-part
     inexact->exact
     inexact?
     input-port?
     integer->char
     integer?
     interaction-environment
     lambda
     lcm
     length
     let
     let*
     let-syntax
     letrec
     letrec-syntax
     list
     list->string
     list->vector
     list-ref
     list-tail
     list?
     load
     log
     magnitude
     make-polar
     make-rectangular
     make-string
     make-vector
     map
     max
     member
     memq
     memv
     min
     modulo
     negative?
     newline
     not
     null-environment
     null?
     number->string
     number?
     numerator
     odd?
     open-input-file
     open-output-file
     (or :syntax)
     output-port?
     pair?
     peek-char
     positive?
     procedure?
     (quasiquote :syntax)
     (quote :syntax)
     quotient
     rational?
     rationalize
     read
     read-char
     real-part
     real?
     remainder
     reverse
     round
     scheme-report-environment
     (set! :syntax)
     set-car!
     set-cdr!
     sin
     sqrt
     string
     string->list
     string->number
     string->symbol
     string-append
     string-ci<=?
     string-ci<?
     string-ci=?
     string-ci>=?
     string-ci>?
     string-copy
     string-fill!
     string-length
     string-ref
     string-set!
     string<=?
     string<?
     string=?
     string>=?
     string>?
     string?
     substring
     symbol->string
     symbol?
     (syntax-rules :syntax)
     tan
     truncate
     values
     vector
     vector->list
     vector-fill!
     vector-length
     vector-ref
     vector-set!
     vector?
     with-input-from-file
     with-output-to-file
     write
     write-char?
     zero?
     )))

(define scheme-env (make-top-level-environment))

(define fake-scheme
  (make-implementation-from-environment scheme-env
                                        (lambda () scheme-interface)))

(define scheme-locale
  (make-locale (lambda () (list fake-scheme))
               (delay (cons eval scheme-env))
               #f
	       #f
               '()))

(define scheme
  (make-implementation-from-locale scheme-locale
                                   (lambda () scheme-interface)))

(define user-locale
  (make-locale (lambda () (list scheme))
               (delay (cons eval scheme-env))
               #f
	       #f
               '()))

(define system-global-environment scheme-locale)

(define user-initial-environment user-locale)

(define import
  (lambda (environment name)
    (environment-link-name (nearest-repl/environment)
                           (->environment environment)
                           name)))

(import '(RUNTIME PARSER) 'LAMBDA-OPTIONAL-TAG)
(import '(RUNTIME PARSER) 'LAMBDA-REST-TAG)

(define (subpackage parent name)
  (let ((parent (find-package parent)))
    (package/add-child! parent
                        name
                        (extend-top-level-environment
                         (package/environment parent)))))

(define config-package
  (subpackage '(USER) 'CONFIG))

(define config-env
  (package/environment config-package))

(eval `(DEFINE (,(string->symbol ".MAKE-SYNTACTIC-TOWER.")
                LOCALE)
         (DELAY (CONS (LAMBDA (FORM LOCALE)
                        (EXTENDED-SCODE-EVAL (SYNTAX FORM LOCALE)
                                             (LOCALE/ENVIRONMENT LOCALE)))
                      (OR LOCALE USER-LOCALE))))
      config-env)

(define (config-ref name)
  (environment-lookup config-env name))

(define (get-locale name)
  (implementation/locale (config-ref name)))
