#| -*-Scheme-*-

$Id$

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

|#

;;;; Module System Syntax

(declare (usual-integrations))

(define-syntax locale
  (er-macro-transformer
   (lambda (form rename compare)
     (let loop ((clauses (cdr form))
                (imports '())
                (tower-clauses '())
		(source-pathname-defaults
		 `(,(rename 'PATHNAME-NEW-TYPE)
		   (,(rename 'DIRECTORY-PATHNAME)
		    (,(rename 'CURRENT-LOAD-PATHNAME)))
		   "scm"))
		(object-pathname-defaults #f)
                (other-clauses '()))
       (cond ((null? clauses)
              `(,(rename 'MAKE-LOCALE)
		(,(rename 'LAMBDA) ()
		  (,(rename 'LIST) ,@(reverse! imports)))
		(,(rename (string->symbol ".MAKE-SYNTACTIC-TOWER."))
		 ,(if (pair? tower-clauses)
		      `(,(rename 'LOCALE) ,@(reverse! tower-clauses))
		      `(,(rename 'QUOTE) #F)))
		,source-pathname-defaults
		,(cond (object-pathname-defaults)
		       ((assq 'FILES other-clauses)
			=> (lambda (clause)
			     (and (= 2 (length clause))
				  `(,(rename 'MERGE-PATHNAMES)
				    ,(cadr clause)
				    (,(rename 'PATHNAME-NEW-TYPE)
				     (,(rename 'CURRENT-LOAD-PATHNAME))
				     (,(rename 'QUOTE) #F))))))
		       (else
			`(,(rename 'QUOTE) #F)))
		(,(rename 'QUOTE) ,(reverse! other-clauses))))
             ((not (pair? clauses))
              (ill-formed-syntax form))
             (else
              (case (caar clauses)
                ((IMPORT)
                 (loop (cdr clauses)
                       (append-reverse (cdar clauses) imports)
                       tower-clauses
		       source-pathname-defaults
		       object-pathname-defaults
                       other-clauses))
                ((FOR-SYNTAX)
                 (loop (cdr clauses)
                       imports
                       (append-reverse (cdar clauses) tower-clauses)
		       source-pathname-defaults
		       object-pathname-defaults
                       other-clauses))
		((SOURCE-PATHNAME-DEFAULTS)
		 (loop (cdr clauses)
		       imports
		       tower-clauses
		       (cadar clauses)
		       object-pathname-defaults
		       other-clauses))
		((OBJECT-PATHNAME-DEFAULTS)
		 (loop (cdr clauses)
		       imports
		       tower-clauses
		       source-pathname-defaults
		       (cadar clauses)
		       other-clauses))
                (else
                 (loop (cdr clauses)
                       imports
                       tower-clauses
		       source-pathname-defaults
		       object-pathname-defaults
                       (cons (car clauses) other-clauses))))))))))

(define-syntax define-implementation
  (syntax-rules ()
    ((DEFINE-IMPLEMENTATION name interface
         (implementation-clause ...)
       locale-clause
       ...)
     (DEFINE name                       ;++ Associate name somehow.
       (IMPLEMENTATION interface (implementation-clause ...)
         locale-clause
         ...)))))

(define-syntax implementation
  (syntax-rules (TRANSLATE)
    ((IMPLEMENTATION interface
         ((TRANSLATE command ...)       ; Only one implementation
          ...)                          ; clause, for now.
       locale-clause ...)
     (MAKE-IMPLEMENTATION-FROM-LOCALE
      (LOCALE locale-clause ...)
      (LAMBDA () interface)
      ;; '(command ... ...)        ; This should work!  Mumble.
      (MAKE-TRANSLATION (APPEND '(command ...) ...))))))

(define-syntax implementations
  (syntax-rules (TRANSLATE)
    ((IMPLEMENTATIONS ((interface (TRANSLATE command ...) ...)
                       ...)
       locale-clause
       ...)
     (LET ((THE-LOCALE (LOCALE locale-clause ...)))
       (VALUES (MAKE-IMPLEMENTATION-FROM-LOCALE
                THE-LOCALE
                (LAMBDA () interface)
                (MAKE-TRANSLATION (APPEND '(command ...) ...)))
               ...)))))

(define-syntax define-interface
  (syntax-rules ()
    ((DEFINE-INTERFACE name interface)
     (DEFINE name interface))))

(define-syntax export
  (syntax-rules ()
    ((EXPORT specifier ...)
     (MAKE-SIMPLE-INTERFACE '(specifier ...)))))

(define-syntax compound-interface
  (syntax-rules ()
    ((COMPOUND-INTERFACE interface ...)
     (MAKE-COMPOUND-INTERFACE (LAMBDA () interface)
                              ...))))

(define-syntax modify
  (syntax-rules ()
    ((MODIFY implementation command ...)
     (MAKE-MODIFIED-IMPLEMENTATION implementation '(command ...)))))
