#| -*-Scheme-*-

$Id$

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

|#

;;;; Name Translations

(declare (usual-integrations))

(define-structure (translation
                   (constructor %make-translation)
                   (conc-name translation/))
  (mapper #f read-only #t)
  (walker #f read-only #t))

(define identity-walker
  (lambda (total-walker singular-processor)
    singular-processor
    (total-walker (lambda (name procedure)
                    (procedure name)))))

(define null-translation
  (%make-translation identity-procedure identity-walker))

(define (translate-name translation name)
  ((translation/mapper translation) name))

(define (walk-translation translation
          total-walker
          singular-processor)
  ((translation/walker translation)
   total-walker
   singular-processor))

(define (make-translation commands)
  (guarantee-list-of-type commands
                          translation-command?
                          "translation command"
                          'MAKE-TRANSLATION)
  (%make-translation (reduce-map compose
                                 process-mapping-command
                                 identity-procedure
                                 commands)
                     (fold-left process-walking-command
                                identity-walker
                                commands)))

(define (compose f g) (lambda (x) (f (g x))))

(define (reduce-map combiner mapper right-identity list)
  (if (null? list)
      right-identity
      (fold-map combiner mapper (mapper (car list)) (cdr list))))

(define (fold-map combiner mapper state list)
  (let loop ((state state) (list list))
    (if (pair? list)
        (loop (combiner (mapper (car list)) state)
              (cdr list))
        state)))

(define (translation-command? datum)
  (and (pair? datum)
       ((case (car datum)
          ((PREFIX) prefix-arguments?)
          ((EXPOSE) expose-arguments?)
          ((HIDE)     hide-arguments?)
          ((ALIAS)   alias-arguments?)
          ((RENAME) rename-arguments?)
          (else false-procedure))
        (cdr datum))))

(define (prefix-arguments? arguments)
  (and (pair? arguments)
       (symbol? (car arguments))
       (null? (cdr arguments))))

(define (expose-arguments? arguments)
  (list-of-type? arguments symbol?))

(define (hide-arguments? arguments)
  (expose-arguments? arguments))

(define (alias-arguments? arguments)
  (list-of-type? arguments
    (lambda (entry)
      (and (pair? entry)
           (symbol? (car entry))
           (pair? (cdr entry))
           (symbol? (cadr entry))
           (null? (cddr entry))))))

(define (rename-arguments? arguments)
  (alias-arguments? arguments))

(define (process-mapping-command command)
  (apply (case (car command)
           ((PREFIX) prefix-mapper)
           ((EXPOSE) expose-mapper)
           ((HIDE)     hide-mapper)
           ((ALIAS)   alias-mapper)
           ((RENAME) rename-mapper))
         (cdr command)))

(define (process-walking-command walker command)
  (apply (case (car command)
           ((PREFIX) prefix-walker)
           ((EXPOSE) expose-walker)
           ((HIDE)     hide-walker)
           ((ALIAS)   alias-walker)
           ((RENAME) rename-walker))
         walker
         (cdr command)))

(define (prefix-mapper prefix)
  (let* ((prefix (symbol-name prefix))
         (prefix-length (string-length prefix)))
    (lambda (name)
      (let ((name (symbol-name name)))
        (and (string-prefix? prefix name)
             (string->symbol (string-tail name prefix-length)))))))

(define (expose-mapper . names)
  (lambda (name)
    (and (memq name names)
         name)))

(define (hide-mapper . names)
  (lambda (name)
    (and (not (memq name names))
         name)))

(define (alias-mapper . internal->external)
  (let ((external->internal (map reverse internal->external)))
    (lambda (name)
      (cond ((assq name external->internal)
             => cadr)
            (else name)))))

(define (rename-mapper . internal->external)
  (let ((external->internal (map reverse internal->external)))
    (lambda (name)
      (cond ((assq name external->internal) => cadr)
            ((assq name internal->external) #f)
            (else name)))))

(define (prefix-walker walker prefix)
  (lambda (total-walker singular-processor)
    (walker (lambda (walk-aliases)
              (total-walker
               (lambda (name procedure)
                 (walk-aliases name
                               (lambda (name)
                                 (procedure
                                  (symbol-append prefix name)))))))
            (lambda (external internal)
              (singular-processor (symbol-append prefix external)
                                  internal)))))

(define (expose-walker walker . names)
  (lambda (total-walker singular-processor)
    total-walker
    (walker (lambda (walk-aliases)
              (for-each (lambda (internal)
                          (walk-aliases
                           internal
                           (lambda (external)
                             (singular-processor external
                                                 internal))))
                        names))
            (lambda (external internal)
              (if (memq external names)
                  (singular-processor external internal))))))

(define (hide-walker walker . names)
  (lambda (total-walker singular-processor)
    (walker (lambda (walk-aliases)
              (total-walker (lambda (name procedure)
                              (walk-aliases name
                                            (lambda (name)
                                              (if (not (memq name names))
                                                  (procedure name)))))))
            (lambda (external internal)
              (if (not (memq external names))
                  (singular-processor external internal))))))

(define (alias-walker walker . internal->external)
  (lambda (total-walker singular-processor)
    (walker (lambda (walk-aliases)
              (total-walker
               (lambda (name procedure)
                 (walk-aliases name procedure)
                 (cond ((assq name internal->external)
                        => (lambda (entry)
                             (walk-aliases (cadr entry)
                                           procedure)))))))
            (lambda (external internal)
              (singular-processor external internal)
              (cond ((assq external internal->external)
                     => (lambda (entry)
                          (singular-processor (cadr entry)
                                              internal))))))))

(define (rename-walker walker . internal->external)
  (lambda (total-walker singular-processor)
    (walker (lambda (walk-aliases)
              (total-walker
               (lambda (name procedure)
                 (cond ((assq name internal->external)
                        => (lambda (entry)
                             (walk-aliases (cadr entry) procedure)))
                       (else (walk-aliases name procedure))))))
            (lambda (external internal)
              (cond ((assq external internal->external)
                     => (lambda (entry)
                          (singular-processor (cadr entry)
                                              internal)))
                    (else (singular-processor external internal)))))))

;;; Edwin Variables:
;;; lisp-indent/walk-translation: 1
;;; End:
